/*
	(C) 1995-99 AROS - The Amiga Research OS
	$Id: obtaingirport.c,v 1.1.1.1 2005/03/15 15:54:21 laire Exp $
 
	Desc: Intuition Function ObtainGIRPort()
	Lang: english
*/
#include <proto/graphics.h>
#include <proto/layers.h>
#include <proto/exec.h>
#include <graphics/rpattr.h>
#include "intuition_intern.h"

/*****************************************************************************
 
	NAME */
#include <graphics/rastport.h>
#include <intuition/cghooks.h>
#include <proto/intuition.h>

AROS_LH1(struct RastPort *, ObtainGIRPort,

		 /*  SYNOPSIS */
		 AROS_LHA(struct GadgetInfo *, gInfo, A0),

		 /*  LOCATION */
		 struct IntuitionBase *, IntuitionBase, 93, Intuition)

/*  FUNCTION
	This function sets up a RastPort for exclusive use by custom
	gadget hook routines. Call this function each time a hook
	routine needs to render into the gadget and ReleaseGIRPort()
	immediately afterwards.
 
	INPUTS
	gInfo - Pointer to GadgetInfo structure, as passed to each
	custom gadget hook function.
 
	RESULT
	Pointer to a RastPort you can render to. NULL if you aren't
	allowed to render into this gadget.
 
	NOTES
	If a routine passes a RastPort, eg. GM_RENDER, ObtainGIRPort()
	needn't be called.
 
	EXAMPLE
 
	BUGS
 
	SEE ALSO
	ReleaseGIRPort()
 
	INTERNALS
 
	HISTORY
	29-10-95    digulla automatically created from
			    intuition_lib.fd and clib/intuition_protos.h
 
*****************************************************************************/
{
	AROS_LIBFUNC_INIT
	AROS_LIBBASE_EXT_DECL(struct IntuitionBase *,IntuitionBase)

	struct ExtRastPort *xrp;

	DEBUG_OBTAINGIRPORT(dprintf("ObtainGIRPort: GInfo 0x%lx Task 0x%lx <%s>\n",
	                            gInfo, FindTask(NULL),
	                            FindTask(NULL)->tc_Node.ln_Name ? FindTask(NULL)->tc_Node.ln_Name : "NULL"));

	if (gInfo && gInfo->gi_RastPort)
	{
		xrp = CloneRastPortContext (IntuitionBase, gInfo->gi_RastPort);
		if (xrp)
		{
			if (xrp->rport.Layer) LockLayerInfo(xrp->rport.Layer->LayerInfo);
			LOCKGADGET
			/*bug("+++++++++++++ OBTAIN rp=%p lay=%p font=%p origfont=%p\n",rp,rp->Layer,rp->Font,gInfo->gi_RastPort->Font);*/

			if (xrp->rport.Layer)
			{
				LockLayer(0,xrp->rport.Layer);

				xrp->context.scroll_x = xrp->rport.Layer->Scroll_X;
				xrp->context.scroll_y = xrp->rport.Layer->Scroll_Y;

				xrp->rport.Layer->Scroll_X = 0;
				xrp->rport.Layer->Scroll_Y = 0;

				xrp->context.clipregion = InstallClipRegion(xrp->rport.Layer, NULL);

				SetWriteMask(&xrp->rport, 0xFF);
				SetRPAttrs(&xrp->rport,RPTAG_DrMd,JAM1,RPTAG_PenMode,TRUE,TAG_DONE);
			}
		}

		DEBUG_OBTAINGIRPORT(dprintf("ObtainGIRPort: RPort 0x%lx\n", xrp));
		return (struct RastPort *)xrp;
	}
	else
	{
		return NULL;
	}

	AROS_LIBFUNC_EXIT
} /* ObtainGIRPort */
